"""Execution task handler."""

import threading
from concurrent.futures import ThreadPoolExecutor, wait
from typing import NamedTuple

import opentaskpy.otflogging
from opentaskpy.remotehandlers.remotehandler import RemoteHandler
from opentaskpy.taskhandlers.taskhandler import TaskHandler


class DefaultProtocolCharacteristics(NamedTuple):
    """Class defining the configuration for default protocols."""

    module: str
    class_: str


TASK_TYPE = "E"


class Execution(TaskHandler):
    """Execution task handler."""

    remote_handlers: list[RemoteHandler] | None = None
    overall_result: bool = False

    _protocol_classes: dict[str, type] = {}  # Class-level cache
    _protocol_lock = threading.Lock()  # Lock for class-level cache

    def __init__(self, global_config: dict, task_id: str, execution_definition: dict):
        """Initialize the execution handler.

        Args:
            global_config (dict): The global config.
            task_id (str): The task ID.
            execution_definition (dict): The execution definition.
        """
        self.task_id = task_id
        self.execution_definition = execution_definition

        self.logger = opentaskpy.otflogging.init_logging(
            "opentaskpy.taskhandlers.execution", self.task_id, TASK_TYPE
        )

        super().__init__(global_config)

    def return_result(
        self,
        status: int,
        message: str | None = None,
        exception: Exception | None = None,
    ) -> bool:
        """Return the result of the task run.

        Args:
            status (int): The status code to return.
            message (str, optional): The message to return. Defaults to None.
            exception (Exception, optional): The exception to return. Defaults to None.

        Returns:
            bool: The result of the task run.
        """
        # Delete the remote connection objects
        if self.remote_handlers:
            for remote_handler in self.remote_handlers:
                remote_host = self._get_remote_host_name(remote_handler)

                self.logger.log(
                    12,
                    f"[{remote_host}] Closing source connection for {remote_handler}",
                )
                remote_handler.tidy()

        # Call super to do the rest
        return super().return_result(status, message, exception)  # type: ignore[no-any-return]

    def _get_remote_host_name(self, remote_handler: RemoteHandler) -> str:
        return (
            str(remote_handler.remote_host)
            if hasattr(remote_handler, "remote_host")
            else "REMOTE"
        )

    def _set_remote_handlers(self) -> None:
        """Set the remote handlers.

        Determine which protocols are in use and create the appropriate objects for
        each.
        """
        # Based on the transfer definition, determine what to do first
        # Based on the remote protocol pick the appropriate remote handler
        # For each host, create a remote handler
        self.remote_handlers = []
        remote_protocol = self.execution_definition["protocol"]["name"]
        self.execution_definition["task_id"] = self.task_id

        if remote_protocol in super().DEFAULT_PROTOCOL_MAP[TASK_TYPE]:
            if "hosts" in self.execution_definition:
                for host in self.execution_definition["hosts"]:
                    handler_class = super()._get_default_class(
                        TASK_TYPE, remote_protocol
                    )
                    remote_handler = handler_class(host, self.execution_definition)

                    self.remote_handlers.append(remote_handler)
            else:
                handler_class = super()._get_default_class(TASK_TYPE, remote_protocol)
                remote_handler = handler_class(self.execution_definition)

                self.remote_handlers.append(remote_handler)
        else:
            remote_handler = super()._get_handler_for_protocol(
                remote_protocol, self.execution_definition
            )
            self.remote_handlers.append(remote_handler)

            super()._set_handler_vars(remote_protocol, remote_handler)

    def run(self, kill_event: threading.Event | None = None) -> bool:
        """Run the execution.

        Args:
            kill_event (threading.Event, optional): An event to kill the execution. Defaults to None.

        Returns:
            bool: The result of the execution.
        """
        self.logger.info("Running execution")

        self._set_remote_handlers()

        # Check that we have remote handlers, if not something has gone very wrong
        if not self.remote_handlers:
            self.logger.error("No remote handlers set")
            return self.return_result(1, "No remote handlers set")

        # This is where we could potentially be waiting for a while. So,
        # for each remote handler, we should spawn a new thread to run
        # the command. Then we can wait for all threads to complete.

        ex = None

        with ThreadPoolExecutor(len(self.remote_handlers)) as executor:
            futures = [
                executor.submit(self._execute, remote_handler)
                for remote_handler in self.remote_handlers
            ]
            self.logger.debug("Triggered all threads")

            while True:
                try:
                    # Sleep 5 seconds for each loop
                    wait(futures, timeout=5)
                    self.logger.info("Waiting for threads to complete...")

                except TimeoutError:
                    pass

                if kill_event and kill_event.is_set():
                    self.logger.info("Kill event received, stopping threads")

                    # Before we terminate everything locally, we need to make sure that the processes
                    # on the remote hosts are terminated as well
                    for remote_handler in self.remote_handlers:
                        remote_host = self._get_remote_host_name(remote_handler)
                        self.logger.info(f"Killing remote processes on {remote_host}")
                        remote_handler.kill()

                    executor.shutdown(wait=False)
                    # Make sure all threaeds are dead
                    for future in futures:
                        future.cancel()
                        # Check it's dead
                        if future.running():
                            self.logger.error(
                                f"Thread {future} is still running after kill. Cannot"
                                " kill a running thread. Will have to wait for it to"
                                " complete. Consider altering the plugin so that it"
                                " cannot block."
                            )

                    return self.return_result(
                        1, "Execution(s) failed - Kill signal received"
                    )

                # Break once all threads are done
                if all(future.done() for future in futures):
                    break

            # Check the results
            failures = False
            for future in futures:
                try:
                    result = future.result()
                    if not result:
                        failures = True
                except Exception as e:  # pylint: disable=broad-exception-caught
                    self.overall_result = False
                    ex = e
                    self.logger.error("Thread returned exception")
                    # Ensure we log the exception
                    self.logger.exception(e)
                    failures = True

            if not failures:
                self.overall_result = True

        if self.overall_result:
            return self.return_result(0, "All executions completed successfully")

        return self.return_result(1, "Execution(s) failed", ex)

    def _execute(self, remote_handler: RemoteHandler) -> bool:
        result: bool = remote_handler.execute()
        remote_host = self._get_remote_host_name(remote_handler)
        self.logger.info(f"[{remote_host}] Execution returned {result}")
        return result



"""Dummy Handler.

This module doesn't actually do anything, it's just used for testing the cacheable
variables.
"""

from random import randint

from opentaskpy.config.variablecaching import cache_utils
from opentaskpy.remotehandlers.remotehandler import RemoteTransferHandler


class DummyTransfer(RemoteTransferHandler):
    """Dummy Transfer Handler."""

    TASK_TYPE = "T"

    def __init__(self, spec: dict):
        """Initialise the handler.

        Args:
            spec (dict): The spec for the transfer. This is either the source, or the
            destination spec.
        """
        super().__init__(spec)

        # Pretend that this handler does something, and gets a new accessToken, we need
        # to update the cache with this new value
        self.spec["accessToken"] = randint(1, 100000)

        # If there's cacheable variables, handle them
        if "cacheableVariables" in spec:
            self.handle_cacheable_variables()

    def handle_cacheable_variables(self) -> None:
        """Handle the cacheable variables."""
        # Obtain the "updated" value from the spec
        for cacheable_variable in self.spec["cacheableVariables"]:

            updated_value = self.obtain_variable_from_spec(
                cacheable_variable["variableName"], self.spec
            )

            cache_utils.update_cache(cacheable_variable, updated_value)

    def supports_direct_transfer(self) -> bool:
        """Return False, as a direct transfer is not supported."""
        return False

    def list_files(
        self,
        directory: str | None = None,  # noqa: ARG002
        file_pattern: str | None = None,  # noqa: ARG002
    ) -> dict:
        """Return list of files that match the source definition.

        Args:
            directory (str, optional): The directory to search in. Defaults to None.
            file_pattern (str, optional): The file pattern to search for. Defaults to
            None.

        Returns:
            dict: A dict of files that match the source definition.
        """
        return {}

    def pull_files_to_worker(
        self, files: list[str], local_staging_directory: str  # noqa: ARG002
    ) -> int:
        """Pull files to the worker.

        This is not applicable for a local transfer, since the files are local already.
        The files will not be transferred as they'll just fill up the worker's disk for
        no reason.

        All args are not used because this function literally does nothing.

        Args:
            files (list): A list of files to download.
            local_staging_directory (str): The local staging directory to move the files
            into.

        Returns:
            int: Always returns 0
        """
        return 0

    def push_files_from_worker(
        self, local_staging_directory: str, file_list: dict | None = None
    ) -> int:
        """Not implemented for this handler."""
        raise NotImplementedError

    def transfer_files(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def pull_files(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def move_files_to_final_location(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def handle_post_copy_action(self, files: list[str]) -> int:  # noqa: ARG002
        """Handle the post copy action specified in the config.

        Args:
            files (list[str]): A list of files that need to be handled.

        Returns:
            int: 0 if successful, 1 if not.
        """
        return 0

    def create_flag_files(self) -> int:
        """Create the flag files on the remote host.

        Raises:
            NotImplementedError: This method is not implemented for this handler.
        """
        raise NotImplementedError



"""Config loader."""

import importlib
import importlib.util
import inspect
import json
import os
import sys
from glob import glob
from typing import Any

import jinja2
from jinja2 import meta

import opentaskpy.otflogging
from opentaskpy.exceptions import (
    DuplicateConfigFileError,
    VariableResolutionTooDeepError,
)
from opentaskpy.filters import default_filters

MAX_DEPTH = 5


class LazyResolvedDict(dict):
    """Dictionary wrapper that resolves values only when accessed."""

    def __init__(
        self,
        values: dict[str, Any],
        resolver: Any,
        resolve_lookups: bool = False,
    ) -> None:
        """Initialise the wrapper around a dictionary of unresolved values."""
        super().__init__(values)
        self._resolver = resolver
        self._resolve_lookups = resolve_lookups

    def _wrap(self, value: Any) -> Any:
        if isinstance(value, dict):
            return LazyResolvedDict(
                value,
                self._resolver,
                resolve_lookups=self._resolve_lookups,
            )

        if isinstance(value, list):
            return [self._wrap(item) for item in value]

        return self._resolver(value, resolve_lookups=self._resolve_lookups)

    def __getitem__(self, key: str) -> Any:
        """Return a lazily resolved value for the requested key."""
        return self._wrap(super().__getitem__(key))

    def get(self, key: str, default: Any = None) -> Any:
        """Return a lazily resolved value when the key exists."""
        if key not in self:
            return default

        return self[key]


class ConfigLoader:
    """Class responsible for loading and validating config files."""

    def __init__(self, config_dir: str) -> None:
        """Parse config files and expand variables.

        Args:
            config_dir (str): The path to the config files to load.
        """
        self.logger = opentaskpy.otflogging.init_logging(__name__)
        self.config_dir = config_dir
        self.global_variables: dict = {}
        self.loaded_filters: dict = {}
        self.file_cache: dict[str, str] = {}  # Cache for preloaded files

        self.logger.log(12, f"Looking in {self.config_dir}")

        # Force Jinja2 to log undefined variables
        jinja2.make_logging_undefined(logger=self.logger, base=jinja2.Undefined)
        self.template_env = jinja2.Environment(undefined=jinja2.StrictUndefined)

        self._load_filters(self.template_env.filters)
        self.loaded_filters = self.template_env.filters

        self._load_global_variables()

        self.lazy_load = False
        if os.environ.get("OTF_LAZY_LOAD_VARIABLES", None) == "1":
            self.lazy_load = True

        self._resolve_templated_variables(lazy_load=self.lazy_load)

        # Preload all files in the config directory
        self._preload_files()

    def _preload_files(self) -> None:
        """Preload all JSON and Jinja2 files in the config directory into memory."""
        self.file_cache = {}
        file_patterns = [
            f"{self.config_dir}/**/*.json",
            f"{self.config_dir}/**/*.json.j2",
        ]
        for pattern in file_patterns:
            for file_path in glob(pattern, recursive=True):
                with open(file_path, encoding="utf-8") as file:
                    self.file_cache[file_path] = file.read()
        self.logger.log(12, f"Preloaded {len(self.file_cache)} files into memory.")

    def _load_filters(self, destination: dict) -> None:
        """Load default filters from opentaskpy.filters.default_filters.

        Args:
            destination (dict): The destination dictionary to load the filters into
        """
        # Prevent multiple loads
        if self.loaded_filters:
            destination.update(self.loaded_filters)
            return

        # Check what functions exist in the module
        for name, func in inspect.getmembers(default_filters, inspect.isfunction):
            destination[name] = func
            self.logger.log(12, f"Registered default filter: {name}")

        # Support for custom filters
        filters_dir = f"{self.config_dir}/filters"
        if os.path.isdir(filters_dir):
            for filter_file in glob(f"{self.config_dir}/filters/*.py", recursive=True):
                module_name = f"{os.path.basename(filter_file).replace('.py', '')}"
                spec = importlib.util.spec_from_file_location(module_name, filter_file)
                if spec and spec.loader:
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)
                    # Register all functions in the module as Jinja2 filters
                    for name, func in inspect.getmembers(module, inspect.isfunction):
                        destination[name] = func
                        self.logger.log(
                            12, f"Registered custom filter: {name} from {filter_file}"
                        )
                else:
                    self.logger.log(12, f"Couldn't import custom filter: {filter_file}")

        destination = self.loaded_filters

    def _override_variables_from_env(self, variables: dict, variable_type: str) -> None:
        """Overrides variables with environment variables."""
        for (  # pylint: disable=too-many-nested-blocks
            env_var_name,
            env_var_value,
        ) in os.environ.items():
            if "." in env_var_name:
                key_path = env_var_name.split(".")
                current_dict = variables
                self.logger.log(12, f"Searching for {env_var_name}")
                for i, key in enumerate(key_path):
                    if isinstance(current_dict, dict) and key in current_dict:
                        if i == len(key_path) - 1:
                            # It's the final key, override the value
                            self.logger.info(
                                f"Overriding nested {variable_type} variable '{env_var_name}' with environment variable."
                            )
                            # Check the original type of the variable, if it was an int, then cast it to an int
                            if isinstance(current_dict[key], int):
                                current_dict[key] = int(env_var_value)
                            else:
                                current_dict[key] = env_var_value
                        else:
                            # Traverse deeper
                            current_dict = current_dict[key]
                    else:
                        # The key path does not exist in the dictionary
                        break

            elif env_var_name in variables:
                self.logger.info(
                    f"Overriding {variable_type} variable ({env_var_name}: {variables[env_var_name]}) with environment variable ({env_var_value})"
                )
                # Check the original type of the variable, if it was an int, then cast it to an int
                if isinstance(variables[env_var_name], int):
                    variables[env_var_name] = int(env_var_value)
                else:
                    variables[env_var_name] = env_var_value

    def get_global_variables(self) -> dict:
        """Return the set of global variables that have been assigned via config files.

        This will first check for anything that has been overridden in the environment
        first, and replace those if necessary.

        Returns:
            dict: _description_
        """
        # Before we return the variables, we need to check for any environment variables that match the same name, and are set
        # if this is the case, then we replace the value with the environment variable
        self._override_variables_from_env(self.global_variables, "global")

        return self.global_variables

    def _resolve_lookup_value(self, value: Any, resolve_lookups: bool = True) -> Any:
        """Resolve templated values before they are passed to a lookup plugin."""
        if isinstance(value, dict):
            return {
                key: self._resolve_lookup_value(item, resolve_lookups=resolve_lookups)
                for key, item in value.items()
            }

        if isinstance(value, list):
            return [
                self._resolve_lookup_value(item, resolve_lookups=resolve_lookups)
                for item in value
            ]

        if isinstance(value, str):
            if not resolve_lookups and "lookup(" in value:
                return value

            return self._resolve_templated_variables_from_string(value)

        return value

    def template_lookup(self, plugin: str, **kwargs) -> str:  # type: ignore[no-untyped-def]
        """Lookup function used by Jinja.

        This function is responsible for calling the lookup plugins to evaluate custom
        variables. This imports the necessary Python module and calls the lookup
        function's run method.

        Args:
            plugin (str): The name of the plugin to call
            **kwargs: The arguments to pass to the plugin

        Returns:
            str: The result of the lookup plugin
        """
        self.logger.log(
            11, f"Got call to lookup function {plugin} with kwargs {kwargs}"
        )

        kwargs = {
            key: self._resolve_lookup_value(value) for key, value in kwargs.items()
        }

        # Append the globals to the kwargs
        kwargs["globals"] = LazyResolvedDict(
            self.global_variables,
            self._resolve_lookup_value,
            resolve_lookups=False,
        )

        # Import the plugin if its not already loaded
        if f"opentaskpy.plugins.lookup.{plugin}" not in sys.modules:
            # Check the module is loadable
            try:
                importlib.import_module(f"opentaskpy.plugins.lookup.{plugin}")
            except ModuleNotFoundError:
                self.logger.log(
                    11,
                    (
                        f"Module not found: opentaskpy.plugins.lookup.{plugin}. Looking"
                        " in plugins directory instead"
                    ),
                )

        # If we haven't loaded the plugin yet, then look in the cfg/plugins directory and see if we can find it
        if f"opentaskpy.plugins.lookup.{plugin}" not in sys.modules:
            plugin_path = f"{self.config_dir}/plugins/lookup/{plugin}.py"
            if os.path.isfile(plugin_path):
                spec = importlib.util.spec_from_file_location(
                    f"opentaskpy.plugins.lookup.{plugin}", plugin_path
                )
                if spec and spec.loader:
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)
                sys.modules[f"opentaskpy.plugins.lookup.{plugin}"] = module

        # If we are in noop mode, then don't actually run the plugin
        if "OTF_NOOP" in os.environ and os.environ["OTF_NOOP"] == "true":
            return "noop"

        # Run the run function of the imported module
        return str(
            getattr(  # noqa: B009
                sys.modules[f"opentaskpy.plugins.lookup.{plugin}"], "run"
            )(**kwargs)
        )

    # TASK DEFINITION FIND FILE
    def load_task_definition(self, task_id: str, cache: bool = True) -> dict:
        """Load the task definition from the config directory.

        Args:
            task_id (str): The id of the task to load
            cache (bool, optional): Whether to use the cache or load from disk. Defaults to True.

        Raises:
            DuplicateConfigFileError: Raised if more than one config file is found
            matching the task_id
            FileNotFoundError: Raised if no config file is found matching the task_id

        Returns:
            dict: A dictionary representing the task definition
        """
        if not cache:
            self._preload_files()

        # Search for files matching the task_id in the preloaded cache
        matching_files = [
            path
            for path in self.file_cache
            if path.endswith(f"{task_id}.json") or path.endswith(f"{task_id}.json.j2")
        ]

        if not matching_files or len(matching_files) != 1:
            if len(matching_files) > 1:
                raise DuplicateConfigFileError(
                    f"Found more than one task with name: {task_id}"
                )

            raise FileNotFoundError(f"Couldn't find task with name: {task_id}")

        found_file = matching_files[0]
        self.logger.log(12, f"Found: {found_file}")

        task_definition = self._enrich_variables(found_file)

        # If the task has any variables, we need to check if they're overridden by environment variables
        if "variables" in task_definition:
            self._override_variables_from_env(task_definition["variables"], "task")

        # Check to see if this is not a batch type
        if "type" in task_definition and task_definition["type"] == "batch":
            self.logger.debug("Cannot apply overrides to batch tasks")
            return task_definition

        # Finally, attributes of the task definition can also be overridden by environment variables
        # e.g. OTF_OVERRIDE_TRANSFER_SOURCE_HOSTNAME will override ["source"]["hostname"], we need to handle this
        # The format is OTF_OVERRIDE_<TASK_TYPE>_<ATTRIBUTE>_<ATTRIBUTE>_<ATTRIBUTE>
        # e.g. OTF_OVERRIDE_TRANSFER_SOURCE_HOSTNAME
        # Some special cases need to be handled, e.g. OTF_OVERRIDE_EXECUTION_PROTOCOL!!SOME_ATTRIBUTE
        # This is where the attribute contains the delimiter itself, so a split by _ will not work
        for key, value in os.environ.items():
            if key.startswith("OTF_OVERRIDE_"):
                # Split the key by _ but handle !! special case first
                split_key: list[str] = []
                while "!!" in key:
                    # Split into parts before and after !! from right
                    prefix, property_name = key.rsplit("!!", maxsplit=1)
                    # Add the property name to the start of the split key list
                    split_key.insert(0, property_name)
                    key = prefix
                # Split the prefix normally by _ and remove OTF_OVERRIDE_
                split_key = key.split("_")[3:] + split_key

                self._apply_env_var_overrides_to_task_definition(
                    task_definition, split_key, value
                )

        return task_definition

    def _apply_env_var_overrides_to_task_definition(
        self, task_definition: dict, split_key: list[str], value: str
    ) -> None:
        attribute = split_key[0]

        # Match the attribute, to an attribute in the task_definition, bearing in mind that the case may not match
        # e.g. "source" may be "Source"
        for task_definition_attribute in task_definition:
            if task_definition_attribute.lower() == attribute.lower():
                attribute = task_definition_attribute
                break

        # Check that the attribute exists in the definition, if not just ignore it and move on to the next override
        if attribute not in task_definition:
            return

        if len(split_key) == 1:
            self.logger.info(
                f"Overriding {attribute}: {task_definition[attribute]} with {value}"
            )
            task_definition[attribute] = value
        elif isinstance(task_definition[attribute], list):
            # If the current attribute is a list, then we need to get the index from the next element
            # e.g. ["source"]["files"][0]["filename"]
            index = int(split_key[1])
            self._apply_env_var_overrides_to_task_definition(
                task_definition[attribute][index], split_key[2:], value
            )
        else:
            self._apply_env_var_overrides_to_task_definition(
                task_definition[attribute], split_key[1:], value
            )

    # POPULATE VARIABLES INSIDE TASK DEFINITION
    # AND LOAD ADDITIONAL VARIABLES FROM TASK DEFINITION
    def _enrich_variables(self, task_definition_file: str) -> dict:
        active_task_definition = None
        with open(task_definition_file, encoding="utf-8") as json_file:
            json_content = json_file.read()
            template = self.template_env.from_string(json_content)

            # If the file is a Jinja2 template, then we do not allow additional
            # variables to be defined in the task definition.
            # Check the file extension
            if not task_definition_file.endswith(".j2"):
                # From this, convert it to JSON and pull out the variables key if there is one
                task_definition = json.loads(json_content)
                # Extend or replace any local variables for this task
                if "variables" in task_definition:
                    self.global_variables = (
                        self.global_variables | task_definition["variables"]
                    )

                template = self.template_env.from_string(json_content)

            self._load_filters(template.globals)

            # Define lookup function
            template.globals["lookup"] = self.template_lookup

            # Using the active task definition, resolve any templated variables
            # Find any variables that are templated from the template object
            if self.lazy_load:

                ast = self.template_env.parse(json_content)
                undeclared_variables = meta.find_undeclared_variables(ast)

                self.logger.log(
                    12,
                    f"Found undeclared variables: {undeclared_variables}",
                )

                # For each undeclared variable, resolve it
                for undeclared_variable in undeclared_variables:

                    # Get the value of the variable from global variables, if it exists,
                    # if it doesn't it must be a task variable, so this can be resolved below
                    if undeclared_variable not in self.global_variables:
                        self.logger.log(
                            12,
                            f"Variable {undeclared_variable} not found in global variables, must be a task variable",
                        )
                        continue

                    unresolved_variable = self.global_variables[undeclared_variable]

                    # Check the type of the variable, if it's not a string, then dump it into a JSON string
                    converted_variable = False
                    if not isinstance(unresolved_variable, str):
                        self.logger.debug(
                            "Converted unresolved variable into a RAW string"
                        )
                        unresolved_variable = json.dumps(unresolved_variable)
                        converted_variable = True

                    self.logger.info(f"Resolving variable {undeclared_variable}")
                    evaluated_variable = self._resolve_templated_variables_from_string(
                        unresolved_variable
                    )

                    self.logger.log(
                        12,
                        f"Resolved variable {undeclared_variable}",
                    )

                    # If the variable was not a string, then convert it back to the original type
                    if converted_variable:
                        self.logger.debug("Converting variable back into a JSON object")
                        evaluated_variable = json.loads(evaluated_variable)

                    # Now update the global variables with the resolved value
                    self.global_variables[undeclared_variable] = evaluated_variable

            rendered_template = template.render(self.global_variables)
            active_task_definition = dict(json.loads(rendered_template))
            self.logger.log(
                12,
                f"Evaluated task definition: {opentaskpy.otflogging.redact(json.dumps(active_task_definition))}",
            )

        return active_task_definition

    # READ AND PARSE ACTUAL VARIABLE FILES
    def _load_global_variables(self) -> None:
        global_variables: dict = {}
        variable_configs: list[str] = []

        # See if variables file(s) have been overridden via environment variable
        if "OTF_VARIABLES_FILE" in os.environ:
            new_variables_files = os.environ["OTF_VARIABLES_FILE"].split(",")
            for new_variables_file in new_variables_files:
                self.logger.info(f"Overriding variables file with {new_variables_file}")
                # Validate that the file exists
                if not os.path.isfile(new_variables_file):
                    raise FileNotFoundError(
                        f"Couldn't find variables file: {new_variables_file}"
                    )
                variable_configs.append(new_variables_file)

        else:
            file_types = (".json.j2", ".json")
            for file_type in file_types:
                variable_configs.extend(
                    glob(f"{self.config_dir}/**/variables{file_type}", recursive=True)
                )

            if not variable_configs:
                raise FileNotFoundError(
                    "Couldn't find any variables.(json|json.j2) files under"
                    f" {self.config_dir}"
                )

        for variable_file in variable_configs:
            with open(variable_file, encoding="utf-8") as json_file:
                this_variable_config = json.load(json_file)
                global_variables = global_variables | this_variable_config

        self.global_variables = global_variables

    # RESOLVE ANY VARIABLES THAT USE OTHER VARIABLES IN THE VARIABLE FILES
    def _resolve_templated_variables(self, lazy_load: bool = False) -> None:
        # We need to evaluate the variables themselves, in case there's any recursion
        # Convert the variables to a JSON string which we can process with the jinja2 templater
        current_depth = 0
        previous_render = None

        template = self.global_variables

        variables_template = self.template_env.from_string(json.dumps(template))
        self._load_filters(variables_template.globals)

        # Define lookup function
        variables_template.globals["lookup"] = self.template_lookup

        # Don't do anything if lazy loading is enabled
        if lazy_load:
            return

        evaluated_variables = variables_template.render(template)

        while evaluated_variables != previous_render and current_depth < MAX_DEPTH:
            previous_render = evaluated_variables

            variables_template = self.template_env.from_string(evaluated_variables)
            self._load_filters(variables_template.globals)

            # Define lookup function
            variables_template.globals["lookup"] = self.template_lookup
            evaluated_variables = variables_template.render(
                json.loads(evaluated_variables)
            )

            current_depth += 1
            if current_depth >= MAX_DEPTH:
                self.logger.error(
                    "Reached max depth of recursive template evaluation. Please check"
                    " the task as variable definitions for infinite recursion"
                )
                raise VariableResolutionTooDeepError(
                    "Reached max depth of recursive template evaluation"
                )

        self.global_variables = json.loads(evaluated_variables)

    def _resolve_templated_variables_from_string(self, json_content: str) -> str:
        # This has some code duplication with _resolve_templated_variables, but it's
        # easier to keep them separate for now

        # We need to evaluate the variables themselves, in case there's any recursion
        # Convert the variables to a JSON string which we can process with the jinja2 templater
        current_depth = 0
        previous_render = None

        variables_template = self.template_env.from_string(json_content)
        self._load_filters(variables_template.globals)

        # Define lookup function
        variables_template.globals["lookup"] = self.template_lookup

        evaluated_variables = variables_template.render(self.global_variables)

        while evaluated_variables != previous_render and current_depth < MAX_DEPTH:
            previous_render = evaluated_variables

            variables_template = self.template_env.from_string(evaluated_variables)

            self._load_filters(variables_template.globals)

            # Define lookup function
            variables_template.globals["lookup"] = self.template_lookup
            evaluated_variables = variables_template.render(self.global_variables)

            current_depth += 1
            if current_depth >= MAX_DEPTH:
                self.logger.error(
                    "Reached max depth of recursive template evaluation. Please check"
                    " the task as variable definitions for infinite recursion"
                )
                raise VariableResolutionTooDeepError(
                    "Reached max depth of recursive template evaluation"
                )

        return str(evaluated_variables)



"""Local Handler.

This module contains the local handlers for transfers.
"""

import glob
import logging
import os
import random
import re
import shutil
import stat
import subprocess
from shlex import quote

import opentaskpy.otflogging
from opentaskpy.remotehandlers.remotehandler import (
    RemoteExecutionHandler,
    RemoteTransferHandler,
)

REMOTE_SCRIPT_BASE_DIR: str = "/tmp"  # nosec B108


class LocalTransfer(RemoteTransferHandler):
    """Local Transfer Handler."""

    TASK_TYPE = "T"

    FILE_NAME_DELIMITER = "|||"

    log_watch_start_row = 0

    def __init__(self, spec: dict):
        """Initialise the handler.

        Args:
            spec (dict): The spec for the transfer. This is either the source, or the
            destination spec.
        """
        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )

        # Handle default values
        if "createDirectoryIfNotExists" not in spec:
            spec["createDirectoryIfNotExists"] = False

        super().__init__(spec)

    def supports_direct_transfer(self) -> bool:
        """Return False, as a direct transfer is not supported."""
        return False

    def list_files(
        self, directory: str | None = None, file_pattern: str | None = None
    ) -> dict:
        """Return list of files that match the source definition.

        Args:
            directory (str, optional): The directory to search in. Defaults to None.
            file_pattern (str, optional): The file pattern to search for. Defaults to
            None.

        Returns:
            dict: A dict of files that match the source definition.
        """
        if not directory:
            directory = str(self.spec["directory"])
        if not file_pattern:
            file_pattern = str(self.spec["fileRegex"])

        self.logger.info(
            f"Searching for files in {directory} with pattern {file_pattern}"
        )

        self.logger.log(
            12,
            f"[LOCAL] Searching in {directory} for files with pattern {file_pattern}",
        )

        files = None
        try:
            files = [
                f"{directory}/{f}"
                for f in os.listdir(directory)
                if re.match(file_pattern, f)
            ]
        except FileNotFoundError:
            files = []

        # For each file get the file age and size (in case we need them for file watches)
        result: dict = {}
        for file in files:
            file_stat = os.stat(file)
            modified_time = file_stat.st_mtime
            size = file_stat.st_size
            result[file] = {}
            result[file]["size"] = size
            result[file]["modified_time"] = modified_time
        return result

    def pull_files_to_worker(
        self, files: list[str], local_staging_directory: str  # noqa: ARG002
    ) -> int:
        """Pull files to the worker.

        This is not applicable for a local transfer, since the files are local already.
        The files will not be transferred as they'll just fill up the worker's disk for
        no reason.

        All args are not used because this function literally does nothing.

        Args:
            files (list): A list of files to download.
            local_staging_directory (str): The local staging directory to move the files
            into.

        Returns:
            int: Always returns 0
        """
        return 0

    def push_files_from_worker(
        self, local_staging_directory: str, file_list: dict | None = None
    ) -> int:
        """Push files from the worker to another local directory.

        Moves files from a local staging directory, to another local destination folder.

        Args:
            local_staging_directory (str): The local staging directory to upload the
            files from.
            file_list (dict, optional): A dictionary of files to transfer. Defaults to

        Returns:
            int: 0 if successful, 1 if not.
        """
        destination_directory = quote(self.spec["directory"])

        # Sanitize the destination directory
        destination_directory = quote(destination_directory)

        self.logger.info("[LOCALHOST] Validating destination dir")
        # Check if the local directory exists
        remote_dir_exists = False
        try:
            os.stat(destination_directory)
            remote_dir_exists = True
        except FileNotFoundError:
            self.logger.info(
                f"[LOCALHOST] Destination dir does not exist: {destination_directory}"
            )

        if not remote_dir_exists and not self.spec["createDirectoryIfNotExists"]:
            self.logger.error(
                f"[LOCALHOST] Destination dir does not exist: {destination_directory}"
            )
            return 1

        if not remote_dir_exists and self.spec["createDirectoryIfNotExists"]:
            self.logger.info(
                f"[LOCALHOST] Creating destination directory: {destination_directory}"
            )
            os.makedirs(destination_directory)

        # Transfer the files
        result = 0

        if file_list:
            files = list(file_list.keys())
        else:
            files = glob.glob(f"{local_staging_directory}/*")

        for file in files:
            file_name = os.path.basename(file)
            final_destination = f"{destination_directory}/{file_name}"
            self.logger.info(
                f"[LOCALHOST] Moving file to new location: {final_destination}"
            )

            # Handle any rename that might be specified in the spec
            if "rename" in self.spec:
                rename_regex = self.spec["rename"]["pattern"]
                rename_sub = self.spec["rename"]["sub"]

                file_name = re.sub(rename_regex, rename_sub, file_name)
                final_destination = f"{destination_directory}/{file_name}"

            mode = self.spec.get("mode", None)

            try:
                shutil.copy(file, final_destination)
                if mode:
                    os.chmod(final_destination, int(mode, base=8))

                self.logger.info(
                    f"[LOCALHOST] Copied file {file} to {final_destination}"
                )
            except Exception as ex:  # pylint: disable=broad-exception-caught
                self.logger.error(f"[LOCALHOST] Failed to move file: {ex}")
                result = 1

        return result

    def transfer_files(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def pull_files(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def move_files_to_final_location(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def handle_post_copy_action(self, files: list[str]) -> int:
        """Handle the post copy action specified in the config.

        Args:
            files (list[str]): A list of files that need to be handled.

        Returns:
            int: 0 if successful, 1 if not.
        """
        if self.spec["postCopyAction"]["action"] == "delete":
            # Loop through each file and use the sftp client to delete the files

            for file in files:
                try:
                    self.logger.info(f"[LOCALHOST] Deleting file {file}")
                    os.remove(file)
                    self.logger.info(f"[LOCALHOST] Deleted file {file}")
                except OSError:
                    self.logger.error(
                        f"[LOCALHOST] Could not delete file {file} on source host"
                    )
                    return 1

        if (
            self.spec["postCopyAction"]["action"] == "move"
            or self.spec["postCopyAction"]["action"] == "rename"
        ):
            # Check that the destination directory exists
            move_dir = os.path.dirname(self.spec["postCopyAction"]["destination"])

            try:
                stat_result = os.stat(move_dir)
                # If it exists, then we need to ensure its a directory and not just a file
                if not stat.S_ISDIR(stat_result.st_mode):  # type: ignore[arg-type]
                    self.logger.error(
                        f"[LOCALHOST] Destination directory {move_dir} is"
                        " not a directory on source host"
                    )
                    return 1
            except OSError:
                self.logger.error(
                    f"[LOCALHOST] Destination directory {move_dir} does"
                    " not exist on source host"
                )
                return 1

                # Loop through the files and move them

            for file in files:
                try:
                    # If this is a move, then just move the file
                    if self.spec["postCopyAction"]["action"] == "move":
                        self.logger.info(
                            f"[LOCALHOST] Moving {file} to"
                            f" {self.spec['postCopyAction']['destination']}"
                        )
                        # Get the actual file name
                        file_name = os.path.basename(file)
                        os.rename(
                            file,
                            f"{self.spec['postCopyAction']['destination']}/{file_name}",
                        )
                        self.logger.info(
                            f"[LOCALHOST] Renamed file {file} to {self.spec['postCopyAction']['destination']}/{file_name}"
                        )
                    # If this is a rename, then we need to rename the file
                    if self.spec["postCopyAction"]["action"] == "rename":
                        # Determine the new file name
                        new_file_dir = os.path.dirname(
                            self.spec["postCopyAction"]["destination"]
                        )
                        current_file_name = os.path.basename(file)

                        rename_regex = self.spec["postCopyAction"]["pattern"]
                        rename_sub = self.spec["postCopyAction"]["sub"]

                        new_file_name = re.sub(
                            rename_regex, rename_sub, current_file_name
                        )

                        self.logger.info(
                            f"[LOCALHOST] Renaming {file} to"
                            f" {new_file_dir}/{new_file_name}"
                        )
                        os.rename(file, f"{new_file_dir}/{new_file_name}")
                        self.logger.info(
                            f"[LOCALHOST] Renamed file {file} to {new_file_dir}/{new_file_name}"
                        )
                except OSError as e:
                    self.logger.error(f"[LOCALHOST] Error: {e}")
                    self.logger.error(
                        f"[LOCALHOST] Error moving or renaming file {file}"
                    )
                    return 1

        return 0

    def create_flag_files(self) -> int:
        """Create the flag files on the remote host.

        Returns:
            int: 0 if successful, 1 if not.
        """
        filename = self.spec["flags"]["fullPath"]

        try:
            # Create an empty file at this path
            with open(filename, "w", encoding="utf-8"):
                pass

            # Set permissions on the file to whatever was specified in the spec,
            # otherwise we leave them as is
            # We cannot change ownership without using sudo, so we don't bother
            if "permissions" in self.spec:
                os.chmod(filename, int(self.spec["permissions"], base=8))

            self.logger.info(f"[LOCALHOST] Created flag file: {filename}")

        except OSError as e:
            self.logger.error(f"[LOCALHOST] Error: {e}")
            self.logger.error(f"[LOCALHOST] Error creating flag file: {filename}")
            return 1

        return 0


def log_stdout(str_stdout: str, hostname: str, logger: logging.Logger) -> None:
    """Log the stdout from a remote command in a nice format.

    Args:
        str_stdout (str): The stdout from the remote command
        hostname (str): The hostname of the remote host
        logger (logging.Logger): The logger to use
    """
    for line in str_stdout.splitlines():
        logger.info(f"[{hostname}] REMOTE OUTPUT: {line}")


class LocalExecution(RemoteExecutionHandler):
    """Class to handle local execution of commands."""

    TASK_TYPE = "E"
    ps_regex = r"(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)"
    remote_pid: int

    def __init__(self, spec: dict):
        """Initialise the LocalExecution class.

        Args:
            spec (dict): The specification for the command
        """
        self.remote_host: str = "LOCAL"
        self.random: int = random.randint(
            100000, 999999
        )  # Random number used to make sure when we kill stuff, we always kill the right thing

        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )

        super().__init__(spec)

    def _get_child_processes(self, parent_pid: int, process_listing: list) -> list:
        """Get the child processes of a given PID.

        Args:
            parent_pid (int): The PID of the parent process
            process_listing (list): The list of processes running on the remote host

        Returns:
            list: A list of child PIDs
        """
        children = []
        for line in process_listing:
            match = re.search(self.ps_regex, line.decode("utf-8"))
            if match:
                if int(match.group(3)) == parent_pid:
                    child_pid = int(match.group(2))
                    # Never add PID 1 or 0!
                    if child_pid in (1, 0):
                        continue
                    self.logger.debug(
                        f"[LOCALHOST] Found child process with PID: {child_pid}"
                    )
                    children.append(child_pid)
                    # Recurse to find the children of this child
                    children.extend(
                        self._get_child_processes(child_pid, process_listing)
                    )
        return children

    def kill(self) -> None:
        """Kill the process."""
        self.logger.info("[LOCALHOST] Killing process")

        # We know the top level remote PID, we need to get all the child processes associated with it
        out = subprocess.check_output(["ps", "-ef"])
        process_listing = out.splitlines()

        # Now we have this, parse it, find the parent PID, and then all the children
        children = self._get_child_processes(self.remote_pid, process_listing)
        children.append(self.remote_pid)
        self.logger.info(
            f"[LOCALHOST] Found {len(children)} child processes to kill - {children}"
        )

        # Now we have the list of children, kill them
        command = f"kill {' '.join([str(x) for x in children])}"
        self.logger.info(f"[LOCALHOST] Killing processes with command: {command}")

        # Make sure we're not killing PID 1 or 0
        if command in ("kill 1", "kill 0"):
            self.logger.error("[LOCALHOST] Refusing to kill PID 1 or 0, aborting")
            return

        subprocess.call(command.split(" "))

    def execute(self) -> bool:
        """Execute the command.

        Returns:
            bool: True if the command was executed successfully, False otherwise
        """
        try:
            directory = quote(self.spec["directory"])

            command = (
                f"echo __OTF_TOKEN__$$_{self.random}__; cd {directory} &&"
                f" {self.spec['command']}"
            )

            self.logger.info(f"[LOCALHOST] Executing command: {command}")

            with subprocess.Popen(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=True,  # nosec B602
            ) as process:
                # Log the stdout and stderr
                for line in iter(lambda: process.stdout.readline(2048), ""):  # type: ignore[union-attr]
                    # If we get an empty line, then we're done
                    if not line:
                        break
                    line = line.decode("utf-8")
                    log_stdout(line, "LOCALHOST", self.logger)

                    # Check the line for the token and pull out the PID
                    regex = f"__OTF_TOKEN__(\\d+)_{self.random}__"
                    pid_search = re.search(regex, line)
                    if pid_search:
                        self.remote_pid = int(pid_search.group(1))
                        self.logger.info(
                            f"[LOCALHOST] Found remote PID: {self.remote_pid}"
                        )

                # Get the return code
                remote_rc = process.wait()
                if remote_rc != 0:
                    self.logger.error(
                        f"[LOCALHOST] Got return code {remote_rc} from command"
                    )
                    return False

            return True
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.logger.error(f"[LOCALHOST] Exception caught: {e}")
            return False

    def tidy(self) -> None:
        """Tidy up the remote connection."""



# pylint: skip-file
# ruff: noqa
import os

from pytest_shell import fs

import opentaskpy.otflogging
from opentaskpy.config.loader import ConfigLoader
from opentaskpy.taskhandlers import batch, execution, transfer
from tests.file_helper import *  # noqa: F403
from tests.fixtures.pgp import *  # noqa: F403, F401
from tests.fixtures.ssh_clients import *  # noqa: F403

os.environ["OTF_NO_LOG"] = "1"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"

parallel_encryption_batch = {
    "type": "batch",
    "tasks": [
        {"order_id": 1, "task_id": "encrypt-file-1-sftp"},
        {"order_id": 2, "task_id": "encrypt-file-2-sftp"},
    ],
}

parallel_decryption_batch = {
    "type": "batch",
    "tasks": [
        {"order_id": 1, "task_id": "decrypt-file-1-sftp"},
        {"order_id": 2, "task_id": "decrypt-file-2-sftp"},
    ],
}


def test_batch_parallel_encryption(
    root_dir, env_vars, clear_logs, store_pgp_keys, setup_sftp_keys
):
    # Assert one of the keys randomly
    assert os.path.exists("/tmp/public_key_1.txt")

    config_loader = ConfigLoader("test/cfg")

    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/sftp_1/src/file-encrypt1.txt": {
                    "content": "hellothere!"
                }
            },
            {
                f"{root_dir}/testFiles/sftp_1/src/file-encrypt2.txt": {
                    "content": "hellothere!"
                }
            },
        ]
    )
    assert os.path.exists(f"{root_dir}/testFiles/sftp_1/src/file-encrypt1.txt")
    assert os.path.exists(f"{root_dir}/testFiles/sftp_1/src/file-encrypt2.txt")

    batch_obj = batch.Batch(
        None,
        f"parallel-encryption-batch",
        parallel_encryption_batch,
        config_loader,
    )

    # Run the batch and expect a true status
    assert batch_obj.run()

    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/file-encrypt1.txt.gpg")
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/file-encrypt2.txt.gpg")


def test_batch_parallel_decryption(
    root_dir, env_vars, clear_logs, store_pgp_keys, setup_sftp_keys
):
    # Assert one of the keys randomly
    assert os.path.exists("/tmp/private_key_1.txt")

    config_loader = ConfigLoader("test/cfg")

    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/file-encrypt1.txt.gpg")
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/file-encrypt2.txt.gpg")

    batch_obj = batch.Batch(
        None,
        f"parallel-decryption-batch",
        parallel_decryption_batch,
        config_loader,
    )

    # Run the batch and expect a true status
    assert batch_obj.run()

    # Assert files have been decrypted and transferred back to sftp1
    assert os.path.exists(
        f"{root_dir}/testFiles/sftp_1/dest/file-encrypt1-decrypted.txt"
    )
    assert os.path.exists(
        f"{root_dir}/testFiles/sftp_1/dest/file-encrypt2-decrypted.txt"
    )



"""Abstract task handler class."""

import threading
from abc import ABC, abstractmethod
from importlib import import_module
from logging import Logger
from sys import modules
from typing import NamedTuple

import opentaskpy.otflogging
from opentaskpy.exceptions import UnknownProtocolError
from opentaskpy.remotehandlers.remotehandler import RemoteHandler


class DefaultProtocolCharacteristics(NamedTuple):
    """Class defining the configuration for default protocols."""

    module: str
    class_: str


class TaskHandler(ABC):
    """Abstract task handler class."""

    TASK_TYPE: str = ""

    DEFAULT_PROTOCOL_MAP = {
        "T": {
            "ssh": DefaultProtocolCharacteristics(
                "opentaskpy.remotehandlers.ssh", "SSHTransfer"
            ),
            "sftp": DefaultProtocolCharacteristics(
                "opentaskpy.remotehandlers.sftp", "SFTPTransfer"
            ),
            "email": DefaultProtocolCharacteristics(
                "opentaskpy.remotehandlers.email", "EmailTransfer"
            ),
            "local": DefaultProtocolCharacteristics(
                "opentaskpy.remotehandlers.local", "LocalTransfer"
            ),
            "dummy": DefaultProtocolCharacteristics(
                "opentaskpy.remotehandlers.dummy", "DummyTransfer"
            ),
        },
        "E": {
            "ssh": DefaultProtocolCharacteristics(
                "opentaskpy.remotehandlers.ssh", "SSHExecution"
            ),
            "local": DefaultProtocolCharacteristics(
                "opentaskpy.remotehandlers.local", "LocalExecution"
            ),
        },
    }

    logger: Logger
    overall_result: bool
    handled_exception: bool = False

    _protocol_classes: dict[str, type] = {}  # Class-level cache
    _addon_packages: dict[str, type] = {}  # Class-level cache for addon packages
    _protocol_lock = threading.Lock()  # Lock for class-level cache

    def __init__(self, global_config: dict):
        """Initialize the class."""
        self.global_config = global_config

    def return_result(
        self, status: int, message: str, exception: type[Exception] | None = None
    ) -> bool:
        """Return the result of the task run.

        Args:
            status (int): The return code of the task run. 0 is success, anything else is failure.
            message (str): Message to return.
            exception (Exception, optional): Exception to return. Defaults to None.

        Returns:
            bool: True if the task was successful, False otherwise.
        """
        if message:
            if status == 0:
                self.logger.info(message)
            else:
                self.logger.error(message)

        if status == 0:
            self.overall_result = True

        opentaskpy.otflogging.close_log_file(self.logger, self.overall_result)

        # Throw an exception if we have one
        if exception and not self.handled_exception:
            self.handled_exception = True
            if callable(exception):
                raise exception(message)

            raise Exception(message)  # pylint: disable=broad-exception-raised

        return status == 0

    @abstractmethod
    def _set_remote_handlers(self) -> None: ...

    @abstractmethod
    def run(self) -> bool:
        """Run the task handler.

        Returns:
            bool: True if the task was successful, False otherwise.
        """

    def _set_handler_vars(
        self, source_protocol: str, remote_handler: RemoteHandler
    ) -> None:
        # If remote handler has a set handler vars method, call it and pass in any variables it might want
        if hasattr(remote_handler, "set_handler_vars"):
            self.logger.log(12, f"Setting handler vars for {source_protocol}")

            # Read the protocol specific variables from the global config
            if (
                self.global_config
                and "global_protocol_vars" in self.global_config
                and next(
                    (
                        item
                        for item in self.global_config["global_protocol_vars"]
                        if item["name"] == source_protocol
                    ),
                )
            ):
                protocol_vars = next(
                    (
                        item
                        for item in self.global_config["global_protocol_vars"]
                        if item["name"] == source_protocol
                    ),
                ).copy()
                # Remove "name" from the dict
                del protocol_vars["name"]

                remote_handler.set_handler_vars(protocol_vars)

    def _get_handler_for_protocol(
        self, protocol_name: str, spec: dict
    ) -> RemoteHandler:
        """Get the handler for a protocol.

        Looks for the protocol package to import. An addon must be accessible in the
        PYTHONPATH to be found and loaded.

        Args:
            protocol_name (str): The name of the protocol.
            spec (dict): The spec for the protocol.

        Raises:
            UnknownProtocolError: Raised if the protocol is unknown.

        Returns:
            RemoteHandler: The remote handler for the protocol.
        """
        # Remove the class name from the end of addon_protocol
        addon_package = ".".join(protocol_name.split(".")[:-1])

        if addon_package == "":
            raise UnknownProtocolError(f"Unknown protocol {protocol_name}")

        try:
            addon_class = self._addon_packages[protocol_name]
            self.logger.debug(f"Using cached addon class for {protocol_name}")
            return addon_class(spec)
        except KeyError:
            self.logger.debug(
                f"Addon package {protocol_name} not yet loaded, importing..."
            )
            with self._protocol_lock:
                if protocol_name not in self._addon_packages:
                    try:
                        if addon_package not in modules:
                            self.logger.debug(
                                f"Thread {threading.current_thread().name} importing {addon_package}"
                            )
                            import_module(addon_package)

                        addon_class = getattr(
                            modules[addon_package], protocol_name.split(".")[-1]
                        )
                        self._addon_packages[protocol_name] = addon_class
                    except ModuleNotFoundError as exc:
                        raise UnknownProtocolError(
                            f"Unknown protocol {protocol_name}"
                        ) from exc
                else:
                    self.logger.debug(
                        f"Thread {threading.current_thread().name} addon {protocol_name} already loaded (inside lock)"
                    )
                addon_class = self._addon_packages[protocol_name]

        return addon_class(spec)

    def _get_default_class(self, task_type: str, protocol_name: str) -> type:
        try:
            return self._protocol_classes[protocol_name]  # Fast path when cached
        except KeyError:
            self.logger.debug(
                f"Protocol: {protocol_name} - has not yet been loaded, importing..."
            )
            self.logger.debug(
                f"Thread {threading.current_thread().name} waiting for protocol lock"
            )
            with self._protocol_lock:  # Slow path with lock
                self.logger.debug(
                    f"Thread {threading.current_thread().name} acquired protocol lock"
                )
                if protocol_name not in self._protocol_classes:  # Double-check
                    class_name = self.DEFAULT_PROTOCOL_MAP[task_type][
                        protocol_name
                    ].class_
                    module_name = self.DEFAULT_PROTOCOL_MAP[task_type][
                        protocol_name
                    ].module
                    if module_name not in modules:
                        self.logger.debug(
                            f"Thread {threading.current_thread().name} importing {module_name}"
                        )
                        import_module(module_name)
                    self._protocol_classes[protocol_name] = getattr(
                        modules[module_name], class_name
                    )
                else:
                    self.logger.debug(
                        f"Thread {threading.current_thread().name} protocol {protocol_name} already loaded (inside lock)"
                    )
                self.logger.debug(
                    f"Thread {threading.current_thread().name} releasing protocol lock"
                )
                return self._protocol_classes[protocol_name]



# pylint: skip-file
# ruff: noqa
import os
from copy import deepcopy

import pytest

from opentaskpy.exceptions import FilesDoNotMeetConditionsError
from opentaskpy.taskhandlers import transfer

dummy_task_definition = {
    "type": "transfer",
    "source": {
        "taskId": 123,
        "accessToken": "0",
        "protocol": {"name": "dummy"},
        "cacheableVariables": [
            {
                "variableName": "accessToken",
                "cachingPlugin": "file",
                "cacheArgs": {
                    "file": "/tmp/cacheable_variable.txt",
                },
            }
        ],
    },
}


def test_dummy_transfer_cacheable_invalid_variable_name():
    from opentaskpy.remotehandlers.dummy import DummyTransfer

    # Copy the task definition
    dummy_task_definition_copy = deepcopy(dummy_task_definition)

    dummy_task_definition_copy["source"]["cacheableVariables"][0][
        "variableName"
    ] = "print('something_bad')"

    with pytest.raises(ValueError) as e:
        DummyTransfer(dummy_task_definition_copy["source"])
        # Check the message
    assert (
        "Variable name print('something_bad') is not a valid variable name."
        in e.value.args[0]
    )


def test_dummy_transfer(tmpdir):
    #  The key thing to test is that the access token
    #  is written to the cache file
    from opentaskpy.remotehandlers.dummy import DummyTransfer

    dummy_task_definition_copy = deepcopy(dummy_task_definition)

    dummy_task_definition_copy["source"]["cacheableVariables"][0]["cacheArgs"][
        "file"
    ] = f"{tmpdir}/cacheable_variable.txt"

    dummy_transfer_obj = transfer.Transfer(
        None, "dummy-transfer", dummy_task_definition_copy
    )

    dummy_transfer_obj.run()

    # Check the cache file exists on the filesystem
    assert os.path.exists(f"{tmpdir}/cacheable_variable.txt")

    # Check the cache file has a random number in it now that's not the original

    with open(f"{tmpdir}/cacheable_variable.txt", "r") as f:
        contents = f.read()
        assert contents != "0"
        assert contents.isdigit()



"""Logging module."""

import json
import logging
import os
import re
import threading
from datetime import datetime

OTF_LOG_FORMAT = (
    "%(asctime)s — %(levelname)s - %(name)s - %(filename)s:%(lineno)s [%(threadName)s]"
    " — %(message)s"
)
LOG_DIRECTORY = (
    "logs"
    if os.environ.get("OTF_LOG_DIRECTORY") is None
    else os.environ.get("OTF_LOG_DIRECTORY")
)

SENSITIVE_KEY_REGEXES = [
    r"password",
    r"passwd",
    r"pass",
    r"secret",
    r"token",
    r"key",
    r"private",
    r"credentials",
    r"authori[sz]ation",
]

SENSITIVE_REGEXES = [
    r"-----BEGIN.*PRIVATE KEY-----.*-----END.*PRIVATE KEY-----",
    r"-----BEGIN.*PRIVATE KEY BLOCK-----.*-----END.*PRIVATE KEY BLOCK-----",
]

REDACT_STRING = "********"


class SensitiveDataFilter(logging.Filter):
    """Filter to redact sensitive data from logs."""

    def filter(self, record):
        """Filter the log message.

        Args:
            record (logging.LogRecord): The log record to filter
        """
        record.msg = self.mask_sensitive_data(record.msg)
        return True

    def mask_sensitive_data(self, message):
        """Redact any sensitive data from the log message.

        Args:
            message (str): The log message to redact

        Returns:
            str: The redacted log message
        """
        # if message is not a string, then return it as is
        if not isinstance(message, str):
            return message

        redacted_message = redact(message)

        return redacted_message


class JSONFormatter(logging.Formatter):
    """JSON formatter.

    This will take a log message and output it as JSON rather than plain text

    Args:
        logging (logging.Formatter): The logging formatter to use
    """

    # Add an init that takes task_id
    def __init__(self, task_id: str | None):
        """Initialise the formatter.

        Args:
            task_id (str | None): The Task ID. Defaults to None.
        """
        super().__init__()
        self.task_id = task_id

    def format(self, record):
        """Format the log message as JSON."""
        task_id = self.task_id
        run_id = os.environ.get("OTF_RUN_ID")
        message = record.msg
        json_log_record = {
            "timestamp": datetime.utcnow().isoformat(),
            "level": record.levelname,
            "function": record.funcName,
            "file": f"{record.pathname}:{record.lineno}",
            "thread": record.threadName,
            "logger": record.name,
            "message": message,
            "task_id": task_id,
            "run_id": run_id,
        }
        record.msg = json.dumps(json_log_record)
        return super().format(record)


def _define_log_file_name(task_id: str | None, task_type: str | None) -> str:
    global LOG_DIRECTORY  # pylint: disable=global-statement
    LOG_DIRECTORY = (
        "logs"
        if os.environ.get("OTF_LOG_DIRECTORY") is None
        else os.environ.get("OTF_LOG_DIRECTORY")
    )

    # Set a custom handler to write to a specific file
    # Get the appropriate timestamp for the log file
    prefix = datetime.now().strftime("%Y%m%d-%H%M%S.%f")[:-3]
    if os.environ.get("OTF_LOG_RUN_PREFIX") is not None:
        # This might have already been set by task_run, so pull it
        # in from the environment
        prefix = f"{os.environ.get('OTF_LOG_RUN_PREFIX')}"
    else:
        os.environ["OTF_LOG_RUN_PREFIX"] = prefix

    if task_type:
        task_type = f"_{task_type}"
    else:
        task_type = ""

    directory = f"{LOG_DIRECTORY}"
    if os.environ.get("OTF_RUN_ID") is not None:
        directory = f"{directory}/{os.environ.get('OTF_RUN_ID')}"
        filename = f"{prefix}{task_type}_{task_id}_running.log"
    else:
        if task_id is None:
            task_id = "no_task_id"
        directory = f"{directory}/{task_id}"
        filename = f"{prefix}{task_type}_running.log"

    return f"{directory}/{filename}"


def init_logging(
    name: str,
    task_id: str | None = None,
    task_type: str | None = None,
    level: int = logging.INFO,
    override_root_logger: bool = False,
) -> logging.Logger:
    """Setup a logger with the custom format and output filename.

    Args:
        name (str): The name of the logger, usually the class name, but can also
        reference the thread, or batch task
        task_id (str | None, optional): The Task ID. Defaults to None.
        task_type (str | None, optional): The Task Type, either T for transfer, E for
            execution or B for batch. Defaults to None, or WRAPPER if the task_id is not set yet
        level (int, optional): The logging level. Defaults to logging.INFO.
        override_root_logger (bool, optional): Whether to set up the root logger

    Returns:
        logging.Logger: The logger object used for logging output.
    """
    # Set the log format
    formatter = logging.Formatter(OTF_LOG_FORMAT)

    # If the task_id isn't set yet, then use the env var
    if not task_id:
        task_id = os.getenv("OTF_TASK_ID")

    if not task_type:
        task_type = "WRAPPER"

    # Check if there's a root logger already
    if override_root_logger:
        sfh = logging.StreamHandler()

        formatter = (
            JSONFormatter(task_id)
            if os.environ.get("OTF_LOG_JSON") and os.environ.get("OTF_LOG_JSON") == "1"
            else formatter
        )

        sfh.setFormatter(formatter)

        # Set the root logger
        root_logger = logging.getLogger()
        root_logger.setLevel(level)
        root_logger.addHandler(sfh)

        # Add SensitiveDataFilter to the root logger
        sensitive_filter = SensitiveDataFilter()
        root_logger.addFilter(sensitive_filter)

    # Create a unique logger object for this task
    if not task_id:
        otf_logger = logging.getLogger(f"{name}")
    else:
        otf_logger = logging.getLogger(f"{name}.{task_id}")

    # Add SensitiveDataFilter to the logger
    sensitive_filter = SensitiveDataFilter()
    otf_logger.addFilter(sensitive_filter)

    # Set verbosity
    otf_logger.setLevel(logging.getLogger().getEffectiveLevel())
    # Ensure the logger is at least at INFO level
    if otf_logger.getEffectiveLevel() > logging.INFO:
        otf_logger.setLevel(logging.INFO)

    # If the log level is set in the environment, then use that
    if os.environ.get("OTF_LOG_LEVEL") is not None:
        otf_logger.setLevel(os.environ["OTF_LOG_LEVEL"])

    # If OTF_NO_LOG is set, then don't create the handler
    if os.environ.get("OTF_NO_LOG") is not None or not task_id:
        return otf_logger

    log_file_name = _define_log_file_name(task_id, task_type)

    tfh = TaskFileHandler(log_file_name)

    # Check there are no handlers of this class already with the same baseFilename
    if not any(
        isinstance(handler, TaskFileHandler)
        and handler.baseFilename == tfh.baseFilename
        for handler in otf_logger.handlers
    ):
        otf_logger.addHandler(tfh)
        tfh.setFormatter(formatter)

    otf_logger.debug("Logging initialised")

    return otf_logger


def set_log_file(logger_name: str, task_type: str = None) -> None:
    """Given a logger name, set the handler as appropriate for the current context.

    Args:
        logger_name (str): The name of the logger to set the handler for.
        task_type (str, optional): The task type. Defaults to None.
    """
    _logger = logging.getLogger(logger_name)

    # Add SensitiveDataFilter to the logger
    sensitive_filter = SensitiveDataFilter()
    _logger.addFilter(sensitive_filter)

    task_id = os.environ.get("OTF_TASK_ID")

    log_file_name = _define_log_file_name(
        task_id, task_type if task_type else "WRAPPER"
    )

    formatter = logging.Formatter(OTF_LOG_FORMAT)

    formatter = (
        JSONFormatter(task_id)
        if os.environ.get("OTF_LOG_JSON") and os.environ.get("OTF_LOG_JSON") == "1"
        else formatter
    )

    tfh = TaskFileHandler(log_file_name)

    # Check there are no handlers of this class already with the same baseFilename
    if not any(
        isinstance(handler, TaskFileHandler)
        and handler.baseFilename == tfh.baseFilename
        for handler in _logger.handlers
    ):
        _logger.addHandler(tfh)
        tfh.setFormatter(formatter)


def get_latest_log_file(task_id: str, task_type: str) -> str | None:
    """Get the latest log file for this task.

    Args:
        task_id (str): The Task ID
        task_type (str): The task type

    Returns:
        str | None: Either the latest log file name, or None if there are no previous
        logs
    """
    log_file_name = _define_log_file_name(task_id, task_type)
    # Obviously the date/time in the filename needs to be replaced with the latest
    # log file
    # Replace the prefix with a regex wildcard
    log_file_name = log_file_name.replace(os.environ["OTF_LOG_RUN_PREFIX"], ".*")
    # Also, we don't want to limit to running jobs, only failed or successful ones
    log_file_name = log_file_name.replace("_running", "(_failed)*")

    logger.debug(f"Looking for log file: {log_file_name}")
    if not os.path.exists(os.path.dirname(log_file_name)):
        return None

    # List the contents of the directory
    log_files = os.listdir(os.path.dirname(log_file_name))
    logger.debug(f"Found {len(log_files)} files")
    logger.debug(f"Log files: {log_files}")
    # Filter the list to only include files that match the log_file_name, and contain a valid date/time prefix
    log_files = [
        f
        for f in log_files
        if re.match(os.path.basename(log_file_name), f)
        and re.match(r"\d{8}-\d{6}\.\d{3}", f.split("_")[0])
    ]
    logger.debug(
        f"Log files after filtering out logs that dont have valid date prefix and {log_file_name}: {log_files}"
    )

    # Unless another date is given, only look at today's logs
    batch_resume_date = datetime.now().date()
    if os.environ.get("OTF_BATCH_RESUME_LOG_DATE"):
        batch_resume_date = datetime.strptime(
            os.environ.get("OTF_BATCH_RESUME_LOG_DATE"), "%Y%m%d"
        ).date()

    # Remove any logs that are not from the given date
    log_files = [
        f
        for f in log_files
        if datetime.strptime(f.split("_")[0], "%Y%m%d-%H%M%S.%f").date()
        == batch_resume_date
    ]
    logger.debug(
        f"Log files after filtering out logs that dont match date: {log_files}"
    )

    # Sort the list by the date/time in the filename
    log_files.sort(key=lambda x: datetime.strptime(x.split("_")[0], "%Y%m%d-%H%M%S.%f"))
    logger.debug(f"Log files after sorting: {log_files}")
    # Get the latest log file
    if log_files:
        log_file_name = f"{os.path.dirname(log_file_name)}/{log_files[-1]}"
        logger.info(f"Latest log file: {log_file_name}")
        # If the last log was a failure, return that, otherwise we just start from scratch, so return nothing
        if "_failed" in log_file_name:
            return log_file_name

        logger.info("No failed log file found. Starting from scratch.")

    return None


def _mkdir(path: str) -> None:
    os.makedirs(path, exist_ok=True)


def close_log_file(logger__: logging.Logger, result: bool = False) -> None:
    """Closes the log file handle.

    Closes the log file, and renames the log file using the handler, based on the result
    argument

    Args:
        logger__ (_type_): The logger that needs to be closed.
        result (bool, optional): The return status of the task that was run. Defaults to False.
    """
    log_file_name = None
    # Close the log file
    for handler in logger__.handlers:
        # If its a task file handler, and the log file still exists
        if isinstance(handler, TaskFileHandler) and os.path.exists(
            handler.baseFilename
        ):
            log_file_name = handler.baseFilename

    log_handlers = []

    if log_file_name:
        new_log_filename = None
        if result:
            new_log_filename = log_file_name.replace("_running", "")
        elif result is not None and not result:
            new_log_filename = log_file_name.replace("_running", "_failed")

        # Loop through every logger that exists and has a handler of this filename, and
        # call the close method on it. Only the last one should rename the file
        for logger_ in list(logging.Logger.manager.loggerDict.values()):
            if isinstance(logger_, logging.Logger):
                for handler in logger_.handlers:
                    if (
                        isinstance(handler, TaskFileHandler)
                        and handler.baseFilename == log_file_name
                    ):
                        log_handlers.append(handler)
                        handler.close()

        # Now everything is closed, we can rename the log file
        # If result is True, then rename the file and remove _running from the name
        if new_log_filename:
            # Use lock to prevent race condition when multiple threads rename same file
            with _file_operations_lock:
                # Check file still exists before renaming (another thread may have renamed it)
                if os.path.exists(log_file_name):
                    os.rename(log_file_name, new_log_filename)

            # Change the basename of the handler to match the new filename, in case it
            # wants to log anything else
            for handler in log_handlers:
                handler.baseFilename = new_log_filename


def redact(log_message: str) -> str:
    """Redact any sensitive information from the log message.

    Args:
        log_message (str): The log message to redact

    Returns:
        str: The redacted log message
    """
    # Try and parse the log message as JSON
    try:
        log_object = json.loads(log_message)
        # If it's a dictionary, then we need to redact any sensitive information
        if isinstance(log_object, dict):
            for key in log_object:
                # If the key matches any of the sensitive regexes, then redact it
                for sensitive_regex in SENSITIVE_KEY_REGEXES:
                    if re.search(sensitive_regex, key, re.IGNORECASE | re.DOTALL):
                        log_object[key] = REDACT_STRING

        log_message = json.dumps(log_object)
    except json.JSONDecodeError:
        pass
    except Exception:  # pylint: disable=broad-except
        pass

    # If it's not JSON, or even if it is, then we need to redact any sensitive information
    # Look at the list of sensitive regexes to match and replace any that match
    for sensitive_regex in SENSITIVE_REGEXES:
        if re.search(sensitive_regex, log_message, re.IGNORECASE | re.DOTALL):
            log_message = re.sub(
                sensitive_regex,
                REDACT_STRING,
                log_message,
                flags=re.IGNORECASE | re.DOTALL,
            )

    return log_message


logger = init_logging(__name__)

# Thread lock for directory creation and file operations
_file_operations_lock = threading.Lock()


# mypy: ignore-errors
class TaskFileHandler(logging.FileHandler):
    """Custom file handler to ensure logs get correct naming.

    This class handles the closing of log files. It will rename based on the argument
    passed into the close method.
    """

    def __init__(self, filename, mode="a", encoding=None, delay=True):
        """Create the log directory.

        Overrides the default __init_ to ensure that the path containing the log exists
        before attempting to write to it.

        """
        _mkdir(os.path.dirname(filename))
        logging.FileHandler.__init__(self, filename, mode, encoding, delay)

    def _open(self):
        """Open the log file, ensuring the directory exists.

        Overrides the FileHandler._open to ensure directory creation is thread-safe
        even when delay=True is used. Also uses line buffering for real-time log updates,
        which is important when using network filesystems like EFS.
        """
        # Ensure directory exists with thread safety before opening file
        with _file_operations_lock:
            _mkdir(os.path.dirname(self.baseFilename))
        # Open with line buffering (buffering=1) for real-time log visibility on network filesystems
        return open(
            self.baseFilename,
            self.mode,
            buffering=1,
            encoding=self.encoding,
            errors=self.errors,
        )



# pylint: skip-file
# ruff: noqa
import os

import pytest

from opentaskpy import exceptions
from opentaskpy.taskhandlers import execution
from tests.fixtures.ssh_clients import *  # noqa: F403

os.environ["OTF_NO_LOG"] = "1"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"
os.environ["PYTHONPATH"] = "test/cfg/addons"

# Create a task definition
custom_task_definition = {
    "type": "execution",
    "protocol": {"name": "custom.test_addon.RandomNumberGenerator"},
}


def test_basic_execution():
    execution_obj = execution.Execution(None, "custom-task", custom_task_definition)
    execution_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert (
        execution_obj.remote_handlers[0].__class__.__name__ == "RandomNumberGenerator"
    )

    # Run the execution and expect a true status
    assert execution_obj.run()



"""Email handler to send files via email."""

import glob
import os
import re
import smtplib
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from logging import DEBUG

import opentaskpy.otflogging
from opentaskpy.remotehandlers.remotehandler import RemoteTransferHandler

MAX_OBJECTS_PER_QUERY = 100


class EmailTransfer(RemoteTransferHandler):
    """A remote handler for email transfers."""

    TASK_TYPE = "T"

    def __init__(self, spec: dict):
        """Initialise the handler."""
        self.protocol_vars: dict

        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )

        # Ensure this logs to the same location as this module
        opentaskpy.otflogging.set_log_file("smtplib")

        super().__init__(spec)

    def supports_direct_transfer(self) -> bool:
        """Return whether this handler supports direct transfers."""
        return False

    def set_handler_vars(self, protocol_vars: dict) -> None:
        """Set the handler variables.

        Set any custom variables that are specifically needed for this handler.

        The schema defines the following variables:
        - smtp_server - The SMTP server to use
        - smtp_port - The SMTP port to use
        - sender - The sender of the email
        - credentials - A dictionary containing the following:
            - username - The username used to authenticate
            - password - The password used to authenticate

        Args:
            protocol_vars (dict): The protocol variables.

        """
        self.protocol_vars = protocol_vars

        # Also pull variables that have been set on the spec level
        # Replace/Append anything defined in self.protocol_vars with anything in self.spec["protocol"]
        for key, value in self.spec["protocol"].items():
            self.protocol_vars[key] = value
        # Remove name
        del self.protocol_vars["name"]

    def push_files_from_worker(
        self, local_staging_directory: str, file_list: dict | None = None
    ) -> int:
        """Push files from the worker to the email recipients.

        Args:
            local_staging_directory (str): The local staging directory.
            file_list (dict, optional): A dictionary of files to transfer. Defaults to None.

        Returns:
            int: The result of the transfer.
        """
        result = 0

        if file_list:
            files = list(file_list.keys())
        else:
            files = glob.glob(f"{local_staging_directory}/*")

        # Rename files if required

        # Don't rename the files themselves, because it's unnecessary, we just need
        # to set the new file names in the email attachment
        file_names = []
        if "rename" in self.spec:
            # Handle any rename that might be specified in the spec

            rename_regex = self.spec["rename"]["pattern"]
            rename_sub = self.spec["rename"]["sub"]

            for file in files:
                file_names.append(re.sub(rename_regex, rename_sub, file))
        else:
            file_names = files

        # Get comma separated list of files
        attachment_file_list = ", ".join([file.split("/")[-1] for file in file_names])

        # Add a body to the email
        content_type = (
            "html"
            if "messageContentType" not in self.spec
            or self.spec["messageContentType"] == "text/html"
            else "plain"
        )

        # Determine whether the load the message from a file or from a string
        # Check that message and messageContentFilename are not both defined
        if "message" in self.spec and "messageContentFilename" in self.spec:
            self.logger.error(
                "message and messageContentFilename cannot both be defined!"
            )
            result = 1
            return result

        # Load message from the file, or use the string if defined
        try:
            if "messageContentFilename" in self.spec:
                with open(self.spec["messageContentFilename"], encoding="utf-8") as f:
                    message = f.read()
            else:
                message = self.spec.get(
                    "message",
                    f"{'<p>Please find attached: <b>' if content_type == 'html' else 'Please find attached: '}{attachment_file_list}{'</b></p>' if content_type == 'html' else ''}",
                )
        except (FileNotFoundError, PermissionError, OSError, UnicodeDecodeError) as ex:
            self.logger.error(
                f"Failed to read message content file: {self.spec['messageContentFilename']}"
            )
            self.logger.error(ex)
            result = 1
            return result

        for email_address in self.spec["recipients"]:
            # Create an email message
            msg = MIMEMultipart()

            # Attach the files to the message
            for file, file_name in zip(files, file_names):
                # Strip the directory from the file name
                file_name = file_name.split("/")[-1]
                self.logger.debug(f"Emailing file: {files} to {email_address}")
                try:
                    with open(file, "rb") as file_handle:
                        part = MIMEApplication(file_handle.read(), Name=file_name)
                    # After the file is closed
                    part["Content-Disposition"] = f'attachment; filename="{file_name}"'
                    msg.attach(part)
                except Exception as ex:  # pylint: disable=broad-exception-caught
                    self.logger.error(f"Failed to attach file: {file}")
                    self.logger.error(ex)
                    result = 1

            msg.attach(MIMEText(message, content_type))

            # Set the email subject
            if "subject" in self.spec:
                msg["Subject"] = self.spec["subject"]

            msg["From"] = self.protocol_vars["sender"]

            # Set connection port, default to 587
            smtp_port = 587
            if "smtp_port" in self.protocol_vars:
                smtp_port = self.protocol_vars["smtp_port"]

            # Determine whether to use startTLS on connection
            use_tls = True
            if "use_tls" in self.protocol_vars and not self.protocol_vars["use_tls"]:
                use_tls = False

            # Send the email using a provided SMTP server
            try:
                self.logger.info(f"Sending email to {email_address}")
                smtp = smtplib.SMTP(
                    self.protocol_vars["smtp_server"],
                    port=smtp_port,
                )
                if self.logger.getEffectiveLevel() <= DEBUG:
                    smtp.set_debuglevel(1)
                if use_tls:
                    smtp.starttls()

                # Authenticate (if credentials specified)
                if "credentials" in self.protocol_vars:
                    smtp.login(
                        self.protocol_vars["credentials"]["username"],
                        self.protocol_vars["credentials"]["password"],
                    )

                smtp.sendmail(
                    self.protocol_vars["sender"], email_address, msg.as_string()
                )
                self.logger.info(f"Email sent to {email_address}")
                smtp.quit()

            except Exception as ex:  # pylint: disable=broad-exception-caught
                self.logger.error(f"Failed to send email to {email_address}")
                self.logger.error(ex)
                result = 1

        # Delete the content file if every email send succeeded
        if (
            result == 0
            and "messageContentFilename" in self.spec
            and (
                "deleteContentFileAfterTransfer" not in self.spec
                or self.spec["deleteContentFileAfterTransfer"]
            )
        ):
            try:
                os.remove(self.spec["messageContentFilename"])
                self.logger.debug(
                    f"Deleted content file: {self.spec['messageContentFilename']}"
                )
            except Exception as ex:  # pylint: disable=broad-exception-caught
                self.logger.error(
                    f"Failed to delete content file: {self.spec['messageContentFilename']}"
                )
                self.logger.error(ex)
                result = 1
        elif "messageContentFilename" in self.spec:
            self.logger.debug(
                "Skipping deletion of content file because one or more emails failed to send"
            )

        return result

    def pull_files_to_worker(self, local_staging_directory: str) -> int:
        """Not implemented for this handler."""
        raise NotImplementedError

    def handle_post_copy_action(self, files: list) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def list_files(
        self, directory: str | None = None, file_pattern: str | None = None
    ) -> dict:
        """Not implemented for this handler."""
        raise NotImplementedError

    def move_files_to_final_location(self, files: list) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def pull_files(self, files: list) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def transfer_files(
        self, files: dict, remote_spec: dict, dest_remote_handler: dict | None = None
    ) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError
